import dataclasses
import os
import re
import shutil
from typing import Union, List, Optional, NamedTuple, Tuple

import libcst as cst
from libcst.metadata import ParentNodeProvider

from trubar.utils import walk_files
from trubar.messages import MsgNode, MsgDict
from trubar.config import config


__all__ = ["collect", "translate", "merge", "missing", "template",
           "ReportCritical", "ReportUpdates", "ReportTranslations", "ReportAll"]


NamespaceNode = Union[cst.Module, cst.FunctionDef, cst.ClassDef]
SomeString = Union[cst.SimpleString, cst.FormattedString]


re_single_quote = re.compile(r"(^|[^\\])'")
re_double_quote = re.compile(r'(^|[^\\])"')
re_braced = re.compile(r"{.+}")


class State(NamedTuple):
    node: NamespaceNode
    name: str


def prefix_for_node(node: NamespaceNode):
    if isinstance(node, cst.FunctionDef):
        return "def "
    elif isinstance(node, cst.ClassDef):
        return "class "
    assert isinstance(node, cst.Module)
    return ""


class StringCollector(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (ParentNodeProvider, )

    @classmethod
    def parse_file(cls, fullname):
        collector = cls()
        with open(fullname, encoding=config.encoding) as f:
            tree = cst.metadata.MetadataWrapper(cst.parse_module(f.read()))
            tree.visit(collector)
        return MsgNode(collector.contexts[0])

    def __init__(self):
        super().__init__()
        self.module: Optional[cst.Module] = None
        self.module_name: Optional[str] = None
        # The stack of nodes (module, class, function) corresponding to
        # the element of stack of contexts
        self.function_stack: List[State] = []
        self.contexts: List[MsgDict] = []

    def visit_Module(self, node: cst.Module) -> bool:
        self.module = node
        self.push_context(node, "")
        return True

    def push_context(self, node: NamespaceNode, name=None) -> None:
        if name is None:
            name = f"{prefix_for_node(node)}`{node.name.value}`"
        self.function_stack.append(State(node, name))
        self.contexts.append({})

    def pop_context(self) -> None:
        state = self.function_stack.pop()
        context = self.contexts.pop()
        if context:
            self.contexts[-1][state.name] = MsgNode(context)

    def is_useless_string(self, node: cst.CSTNode) -> bool:
        # This is primarily to exclude docstrings: exclude strings if they
        # represent the entire body of a simple statement.
        # It will not exclude, e.g. line `"a" + "b"`.
        parent = self.get_metadata(ParentNodeProvider, node)
        grand = self.get_metadata(ParentNodeProvider, parent)
        return isinstance(parent, cst.Expr) \
            and isinstance(grand, cst.SimpleStatementLine) \
            and len(grand.body) == 1

    def visit_ClassDef(self, node: cst.ClassDef) -> bool:
        self.push_context(node)
        return True

    def leave_ClassDef(self, _) -> None:
        self.pop_context()

    def visit_FunctionDef(self, node: cst.FunctionDef) -> bool:
        self.push_context(node)
        return True

    def leave_FunctionDef(self, _) -> None:
        self.pop_context()

    def visit_FormattedString(
            self,
            node: cst.FormattedString) -> bool:
        lq = len(node.quote)
        if not self.is_useless_string(node):
            text = self.module.code_for_node(node)[len(node.prefix) + lq:-lq]
            self.contexts[-1][text] = MsgNode(None)
        return False  # don't visit anything within an f-string!

    def visit_SimpleString(self, node: cst.SimpleString) -> bool:
        lq = len(node.quote)
        s = self.module.code_for_node(node)[len(node.prefix) + lq:-lq]
        if s and not self.is_useless_string(node):
            self.contexts[-1][s] = MsgNode(None)
        return False  # doesn't matter, there's nothing down there anyway


class TranslationError(Exception):
    pass


class StringTranslator(cst.CSTTransformer):
    def __init__(self, context: MsgDict, module: cst.Module):
        super().__init__()
        self.module = module
        self.context_stack: List[MsgDict] = [context]

    @property
    def context(self):
        return self.context_stack[-1]

    def _error_context(self):
        return ":".join(
            [name for name, sub in prev.items() if sub is cont][0]
            for prev, cont in zip(self.context_stack, self.context_stack[1:]))

    def push_context(self, node: NamespaceNode) -> None:
        key = f"{prefix_for_node(node)}`{node.name.value}`"
        space = self.context[key].value if key in self.context else {}
        self.context_stack.append(space)

    def pop_context(self) -> None:
        self.context_stack.pop()

    def __leave(self, _, updated_node: cst.CSTNode) -> cst.CSTNode:
        self.pop_context()
        return updated_node

    def __translate(
            self,
            node: SomeString, updated_node: SomeString) -> cst.CSTNode:
        if not self.context:
            return updated_node
        lq = len(node.quote)
        original = self.module.code_for_node(node)[len(node.prefix) + lq:-lq]
        if original not in self.context:
            return updated_node
        translation = self.context[original].value
        if translation in (None, False, True):
            return updated_node
        assert isinstance(translation, str)

        quote = node.quote
        if config.smart_quotes:
            has_single = re_single_quote.search(translation)
            has_double = re_double_quote.search(translation)
            if quote == "'" and has_single and not has_double:
                quote = '"'
            elif quote == '"' and has_double and not has_single:
                quote = "'"

        if config.auto_prefix \
                and "f" not in node.prefix and not re_braced.search(original) \
                and re_braced.search(translation) :
            try:
                new_node = cst.parse_expression(
                    f'f{node.prefix}{quote}{translation}{quote}')
            except cst.ParserSyntaxError:
                pass
            else:
                assert isinstance(new_node, cst.FormattedString)
                if any(isinstance(part, cst.FormattedStringExpression)
                       for part in new_node.parts):
                    return new_node

        try:
            return cst.parse_expression(f'{node.prefix}{quote}{translation}{quote}')
        except cst.ParserSyntaxError:
            if "\n" in translation and len(quote) != 3:
                unescaped = " Unescaped \\n?"
            else:
                unescaped = ""
            raise TranslationError(
                f'\nProbable syntax error in translation.{unescaped}\n'
                f'Original: {original}\n'
                f'Translation: {translation}') from None

    visit_ClassDef = push_context
    visit_FunctionDef = push_context

    leave_ClassDef = __leave
    leave_FunctionDef = __leave

    def leave_FormattedString(
            self,
            original_node: cst.FormattedString,
            updated_node: cst.FormattedString) -> cst.CSTNode:
        return self.__translate(original_node, updated_node)

    def leave_SimpleString(
            self,
            original_node: cst.SimpleString,
            updated_node: cst.SimpleString) -> cst.CSTNode:
        return self.__translate(original_node, updated_node)


def collect(source: str,
            existing: Optional[MsgDict] = None,
            pattern: str = "",
            *, quiet=False, min_time=None) -> Tuple[MsgDict, MsgDict]:
    messages = {}
    removed = {}
    # No pattern when calling walk_files: we must get all files so that
    # existing messages in skipped files are kept. We check the pattern here.
    for name, fullname in walk_files(source, "", select=True):
        if pattern in name and (
                min_time is None or os.stat(fullname).st_mtime >= min_time):
            if not quiet:
                print(f"Parsing {name}")
            collected = StringCollector.parse_file(fullname)
            if collected.value:
                messages[name] = collected
            if name in existing:
                removals = MsgNode(merge(
                    existing.pop(name).value, collected.value,
                    "", name, print_unused=False))
                if removals.value:
                    removed[name] = removals
        elif name in existing:
            messages[name] = existing.pop(name)

    existing = {name: trans for name, trans in existing.items()
                if _any_translations(trans.value)}
    removed.update(existing)
    return messages, removed


ReportCritical, ReportUpdates, ReportTranslations, ReportAll = range(4)


def translate(translations: MsgDict,
              source: str, destination: str, pattern: str,
              *, verbosity=ReportUpdates, dry_run=False) -> None:
    def write_if_different(data, dest):
        try:
            with open(dest, encoding=config.encoding) as f:
                diff = 1 if f.read() != data else 0
        except OSError:
            diff = 2
        if diff and not dry_run:
            with open(dest, "wt", encoding=config.encoding) as f:
                f.write(data)
        return diff

    def copy_if_different(src, dest):
        with open(src, encoding=config.encoding) as f:
            return write_if_different(f.read(), dest)

    inplace = source == destination
    if dry_run or inplace:
        def noop(*_, **_1):
            pass

        copyfile = copytree = makedirs = noop

    else:
        copyfile, copytree = shutil.copyfile, shutil.copytree
        makedirs = os.makedirs

    any_reports = False

    def report(s, level):
        nonlocal any_reports
        if level <= verbosity:
            any_reports = True
            print(s)

    for name, fullname in walk_files(source, pattern, select=False):
        transname = os.path.join(destination, name)
        path, _ = os.path.split(transname)
        makedirs(path, exist_ok=True)

        # Copy anything that is not Python
        if not name.endswith(".py") \
                or config.exclude_re and config.exclude_re.search(fullname):
            copyfile(fullname, transname)
            continue

        # Copy files without translations
        if name not in translations or \
                not _any_translations(translations[name].value):
            if inplace:
                continue
            diff = copy_if_different(fullname, transname)
            if diff:
                report(f"Copying {name} (no translations)", ReportAll)
            else:
                report(f"Skipping {name} (unchanged; no translations)",
                       ReportAll)
            continue

        # Parse original sources
        try:
            with open(fullname, encoding=config.encoding) as f:
                orig_source = f.read()
                tree = cst.parse_module(orig_source)
        except Exception:
            print(f"Error when parsing {name}")
            raise

        # Replace with translations, produce new sources
        try:
            translator = StringTranslator(translations[name].value, tree)
            translated = tree.visit(translator)
            trans_source = tree.code_for_node(translated)
        except Exception:
            print(f"Error when inserting translations into {name}")
            raise
        if config.auto_import:
            trans_source = config.auto_import + "\n\n" + trans_source

        diff = write_if_different(trans_source, transname)
        if diff == 0:
            report(f"Skipping {name} (unchanged)", ReportTranslations)
        elif diff == 1:
            report(f"Updating translated {name}", ReportUpdates)
        else:  # diff == 2
            report(f"Creating translated {name}", ReportUpdates)
    if not dry_run:
        for path in config.static_files:
            report(f"Copying files from '{path}'", ReportAll)
            copytree(path, destination, dirs_exist_ok=True)

    if not any_reports and verbosity > ReportCritical:
        print("No changes.")


def _any_translations(translations: MsgDict):
    return any(isinstance(value, str)
               or isinstance(value, dict) and _any_translations(value)
               for value in (msg.value for msg in translations.values()))


def missing(translations: MsgDict,
            messages: MsgDict,
            pattern: str = "") -> MsgDict:
    no_translations: MsgDict = {}
    for obj, orig in messages.items():
        if pattern not in obj:
            continue
        trans = translations.get(obj)
        if trans is None:
            no_translations[obj] = orig  # orig may be `None` or a whole subdict
        elif isinstance(orig.value, dict):
            if submiss := missing(translations[obj].value, orig.value, ""):
                no_translations[obj] = MsgNode(submiss,
                                               trans.comments or orig.comments)
        elif trans.value is None:
            no_translations[obj] = trans  # this keeps comments
    return no_translations


def merge(additional: MsgDict, existing: MsgDict, pattern: str = "",
          path: str = "", print_unused=True) -> MsgDict:
    unused: MsgDict = {}
    for msg, trans in additional.items():
        if pattern not in msg:
            continue
        npath = path + "/" * bool(path) + msg
        if msg not in existing:
            if trans.value and (not isinstance(trans.value, dict)
                                or _any_translations(trans.value)):
                if print_unused:
                    print(f"{npath} not in target structure")
                unused[msg] = trans
        elif isinstance(trans.value, dict):
            subreject = merge(trans.value, existing[msg].value, "", npath,
                              print_unused=print_unused)
            if subreject:
                unused[msg] = MsgNode(subreject)
        elif trans.value is not None:
            existing[msg] = trans
    return unused


def template(existing: MsgDict, pattern: str = "") -> MsgDict:
    new_template: MsgDict = {}
    for msg, trans in existing.items():
        if pattern not in msg:
            continue
        if isinstance(trans.value, dict):
            if subtemplate := template(trans.value):
                new_template[msg] = MsgNode(subtemplate, trans.comments)
        elif trans.value is not False:
            new_template[msg] = MsgNode(None, trans.comments)
    return new_template


@dataclasses.dataclass
class Stat:
    translated: int = 0
    kept: int = 0
    untranslated: int = 0
    programmatic: int = 0

    def __add__(self, other):
        return Stat(self.translated + other.translated,
                    self.kept + other.kept,
                    self.untranslated + other.untranslated,
                    self.programmatic + other.programmatic)

    def __abs__(self):
        return self.translated + self.kept \
               + self.untranslated + self.programmatic

    @classmethod
    def collect_stat(cls, messages):
        values = [obj.value for obj in messages.values()]
        return sum((cls.collect_stat(value)
                    for value in values if isinstance(value, dict)),
                   start=Stat(sum(isinstance(val, str) for val in values),
                              values.count(True),
                              values.count(None),
                              values.count(False)))


def stat(messages: MsgDict, pattern: str):
    if pattern:
        messages = {k: v for k, v in messages.items() if pattern in k}
    stats = Stat.collect_stat(messages)
    n_all = abs(stats)
    if not n_all:
        print("No messages")
    else:
        print(f"Total messages: {abs(stats)}")
        print()
        print(f"{'Translated:':16}"
              f"{stats.translated:6}{100 * stats.translated / n_all:8.1f}%")
        print(f"{'Kept unchanged:':16}"
              f"{stats.kept:6}{100 * stats.kept / n_all:8.1f}%")
        print(f"{'Programmatic:':16}"
              f"{stats.programmatic:6}{100 * stats.programmatic / n_all:8.1f}%")
        print(f"{'Total completed:':16}"
              f"{n_all - stats.untranslated:6}"
              f"{100 - 100 * stats.untranslated / n_all:8.1f}%")
        print()
        print(f"{'Untranslated:':16}"
              f"{stats.untranslated:6}{100 * stats.untranslated / n_all:8.1f}%")



import sys
import os
import re
import locale
import dataclasses

import yaml


@dataclasses.dataclass
class Configuration:
    smart_quotes: bool = True
    auto_prefix: bool = True

    auto_import: str = ""

    static_files: tuple = ()

    exclude_pattern: str = "tests/test_"

    encoding: str = \
        "locale" if sys.version_info >= (3, 10) \
        else locale.getpreferredencoding(False)

    def __post_init__(self):
        self.__update_exclude_re()

    def update_from_file(self, filename):
        if not os.path.exists(filename):
            print(f"Can't open configuration file {filename}")
            sys.exit(4)
        try:
            with open(filename, encoding=self.encoding) as f:
                settings = yaml.load(f, Loader=yaml.Loader)
        except yaml.YAMLError as exc:
            print(f"Invalid configuration file: {exc}")
            sys.exit(4)

        fieldict = {field.name: field for field in dataclasses.fields(self)}
        for name, value in settings.items():
            name = name.replace("-", "_")
            field = fieldict.get(name, None)
            if field is None:
                print(f"Unrecognized configuration setting: {name}")
                sys.exit(4)
            if field.type is bool and value not in (True, False):
                print(f"Invalid value for '{name}': {value}")
                sys.exit(4)
            else:
                try:
                    if value is None:
                        value = field.type()
                    elif field.type is tuple and isinstance(value, str):
                        value = (value, )
                    else:
                        value = field.type(value)
                except ValueError:
                    print(f"Invalid value for '{name}': {value}")
                    sys.exit(4)
            setattr(self, name, value)

        self.__update_exclude_re()
        if isinstance(self.static_files, str):
            self.static_files = [self.static_files]
        self.__check_static_files()

    def set_static_files(self, static):
        self.static_files = static
        self.__check_static_files()

    def set_exclude_pattern(self, pattern):
        self.exclude_pattern = pattern
        self.__update_exclude_re()

    def __update_exclude_re(self):
        # This function is called from (post)init
        # pylint: disable=attribute-defined-outside-init
        self.exclude_pattern = self.exclude_pattern.strip()
        if self.exclude_pattern:
            self.exclude_re = re.compile(self.exclude_pattern)
        else:
            self.exclude_re = None

    def __check_static_files(self):
        for path in self.static_files:
            if not os.path.exists(path):
                print(f"Static files path '{path}' does not exist")
                sys.exit(4)


config = Configuration()



import re
import io
import os
from copy import deepcopy
import unittest
from unittest.mock import Mock, patch
from contextlib import redirect_stdout

import libcst as cst

from trubar.actions import \
    StringCollector, StringTranslator, Stat, \
    collect, missing, merge, template, TranslationError

import trubar.tests.test_module
from trubar.messages import dict_from_msg_nodes, dict_to_msg_nodes, MsgNode
from trubar.tests import yamlized

test_module_path = os.path.split(trubar.tests.test_module.__file__)[0]


class StringCollectorTest(unittest.TestCase):
    @staticmethod
    def collect(s):
        collector = StringCollector()
        tree = cst.metadata.MetadataWrapper(cst.parse_module(s))
        tree.visit(collector)
        return dict_from_msg_nodes(collector.contexts[0])

    def test_simple_string(self):
        msgs = self.collect("""
def f(x):
    a = "a string"
    b = 'another string'
    c = \"\"\"and yet another\"\"\"
    d = '''and there's more'''
""")
        expected = {'def `f`': {'a string': None,
                                'another string': None,
                                'and yet another': None,
                                "and there's more": None}}
        self.assertEqual(msgs, expected)
        # order should match, too
        self.assertEqual(list(msgs["def `f`"]),
                         list(expected["def `f`"]))

    def test_formatted_string(self):
        msgs = self.collect("""
        
def f(x):
    a = f"a string {x}"
    b = f'another string {2 + 2}'
    c = f\"\"\"and yet another {3 + 3}\"\"\"
    d = f'''and there's more'''
""")
        expected = {'def `f`': {
            'a string {x}': None,
            'another string {2 + 2}': None,
            'and yet another {3 + 3}': None,
            "and there's more": None}}

        self.assertEqual(msgs, expected)
        # order should match, too
        self.assertEqual(list(msgs["def `f`"]),
                         list(expected["def `f`"]))

    def test_class_func(self):
        msgs = self.collect("""
class A:
    def b(x):
        def c(x):
            d = "foo"
            e = "bar"
        a = "baz"
        
        class B:
           f = "baz"
           
class C:
    g = "crux" 
""")

        # check structure and order, thus `repr`
        self.assertEqual(
            repr(msgs),
            "{'class `A`': {'def `b`': {'def `c`': {'foo': None, 'bar': None}, "
            "'baz': None, 'class `B`': {'baz': None}}}, 'class `C`': {'crux': None}}")

    def test_module_and_walk_and_collect(self):
        msgs, _ = collect(test_module_path, {}, "", quiet=True)
        self.assertEqual(
             dict_from_msg_nodes(msgs),
            {
             'bar_module/foo_module/__init__.py': {
                 "I've seen things you people wouldn't believe...": None},
             'bar_module/__init__.py': {
                 'Attack ships on fire off the shoulder of Orion...': None},
             'baz_module/__init__.py': {'def `f`': {
                 'I watched C-beams glitter in the dark near the Tannhäuser Gate.': None}},
             '__init__.py': {'class `Future`': {
                'All those moments will be lost in time, like tears in rain...': None,
                'Time to die.': None}},
            }
        )

    def test_no_docstrings(self):
        msgs = self.collect('''
"""docstring"""

def f(x):
    "docstring"
    a = "not a docstring"

def g(x):
    """Also docstring"""
    f("bar")
    "useless string"''')
        self.assertEqual(
            msgs,
            {'def `f`': {'not a docstring': None},
                  'def `g`': {'bar': None}})

    def test_no_strings_within_interpolation(self):
        msgs = self.collect("""a = f'x = {len("foo")} {"bar"}'""")
        self.assertEqual(msgs, {'x = {len("foo")} {"bar"}': None})


class StringTranslatorTest(unittest.TestCase):
    def test_translator(self):
        module = """
# A comment
class A:
    def b(x):  # Another comment
        def c(x):
              d = '''foo'''  # intentional bad indentation
              e = "bar"
        a = "baz"

        class B:
           f = f"baz{42}"


class C:
    g = 'crux'
"""

        translations = {'class `A`': {'def `b`': {'def `c`': {'foo': 'sea food',
                                                              'bar': None},
                                                  'baz': True,
                                                  'class `B`': {'baz{42}':
                                                                    False}}},
                        'class `C`': {'crux': ""}}

        tree = cst.parse_module(module)
        translator = yamlized(StringTranslator)(translations, tree)
        translated = tree.visit(translator)
        trans_source = tree.code_for_node(translated)
        self.assertEqual(trans_source, """
# A comment
class A:
    def b(x):  # Another comment
        def c(x):
              d = '''sea food'''  # intentional bad indentation
              e = "bar"
        a = "baz"

        class B:
           f = f"baz{42}"


class C:
    g = ''
""")

    def test_auto_quote_and_prefix(self):
        module = """
print("Foo")
print('Bar')
print("2 + 2 = 4")
print("2 + 3 = 4")
print("2 + 4 = 4")
print("baz")
print("kux")
print(r"bing")
print('''f x g''')
print(\"\"\"f y g\"\"\")
print("just {braces}, not an f-string!")
"""

        tree = cst.parse_module(module)
        translations = {"Foo": 'Fo"o',
                        "Bar": "B'ar",
                        "2 + 2 = 4": "2 + 2 = {2 + 2}",
                        "2 + 3 = 4": "2 + 3 = {}",
                        "2 + 4 = 4": "2 + 4 = {not an expression}",
                        "baz": "ba{}z",
                        "kux": "ku{--}x",
                        "bing": "b{2 + 2}ng",
                        "f x g": "f ' g",
                        "f y g": 'f " g',
                        "just {braces}, not an f-string!": "samo {oklepaji}!"
                        }
        translator = yamlized(StringTranslator)(translations, tree)
        translated = tree.visit(translator)
        trans_source = tree.code_for_node(translated)
        self.assertEqual(trans_source, """
print('Fo"o')
print("B'ar")
print(f"2 + 2 = {2 + 2}")
print("2 + 3 = {}")
print("2 + 4 = {not an expression}")
print("ba{}z")
print("ku{--}x")
print(fr"b{2 + 2}ng")
print('''f ' g''')
print(\"\"\"f " g\"\"\")
print("samo {oklepaji}!")
""")

    def test_syntax_error(self):
        tree = cst.parse_module("print('foo')")
        translator = yamlized(StringTranslator)({"foo": 'bar\nbaz'}, tree)
        self.assertRaisesRegex(
            TranslationError,
            re.compile(".*foo.*bar.*", re.DOTALL), tree.visit, translator)


class ActionsTest(unittest.TestCase):
    @patch("builtins.print")
    def test_collect(self, print_):
        def parse_file(fn):
            c = fn[-4]
            return MsgNode(dict_to_msg_nodes(
                {f"def `{c}`": {f"{c}{c}": None,
                                f"{c}{c}{c}": None},
                 f"{c}": None}))
        file_list = [("a.py", "x/a.py"),
                     ("b/c.py", "x/b/c.py"),
                     ("b/d.py", "x/b/d.py")]
        with patch("trubar.actions.StringCollector.parse_file", parse_file), \
                patch("trubar.actions.walk_files", Mock(return_value=file_list)):
            mess, remo = collect("", {}, "", quiet=True)
            self.assertEqual(
                mess,
                dict_to_msg_nodes({
                    "a.py": {"def `a`": {"aa": None, "aaa": None}, "a": None},
                    "b/c.py": {"def `c`": {"cc": None, "ccc": None}, "c": None},
                    "b/d.py": {"def `d`": {"dd": None, "ddd": None}, "d": None},
                })
            )
            self.assertEqual(remo, {})
            print_.assert_not_called()

            collect("", {}, "", quiet=False)
            print_.assert_called()
            print_.reset_mock()

            existing = dict_to_msg_nodes({
                "b/d.py": {"def `d`": {"dd": "dtrans", "ddd": None},
                           "foo": None, "bar": "baz"},
                "b/e.py": {"qux": None},
                "b/f.py": {"qui": "quo"}})
            mess, remo = collect("", deepcopy(existing), "", quiet=True)
            self.assertEqual(
                mess,
                dict_to_msg_nodes({
                    "a.py": {"def `a`": {"aa": None, "aaa": None}, "a": None},
                    "b/c.py": {"def `c`": {"cc": None, "ccc": None}, "c": None},
                    "b/d.py": {"def `d`": {"dd": "dtrans", "ddd": None}, "d": None},
                })
            )
            self.assertEqual(
                remo,
                dict_to_msg_nodes({
                    "b/d.py": {"bar": "baz"},
                    "b/f.py": {"qui": "quo"}
                })
            )
            print_.assert_not_called()

            existing = dict_to_msg_nodes({
                "b/d.py": {"def `d`": {"dd": "dtrans", "ddd": None},
                           "foo": None, "bar": "baz"},
                "b/e.py": {"qux": None},
                "b/f.py": {"qui": "quo"},
                "b/g.py": {"qux": None},
                "b/h.py": {"qui": "quo"},
            })
            file_list += [("b/e.py", "x/b/e.py"),
                          ("b/f.py", "x/b/f.py")]
            mess, remo = collect("", deepcopy(existing), "d", quiet=True)
            self.assertEqual(
                mess,
                dict_to_msg_nodes({
                    "b/d.py": {"def `d`": {"dd": "dtrans", "ddd": None}, "d": None},
                    "b/e.py": {"qux": None},
                    "b/f.py": {"qui": "quo"}
                })
            )
            self.assertEqual(
                remo,
                dict_to_msg_nodes({
                    "b/d.py": {"bar": "baz"},
                    "b/h.py": {"qui": "quo"}
                })
            )
            print_.assert_not_called()

    @patch("builtins.print")
    def test_collect_empty_file(self, _):
        def parse_file(fn):
            if fn == "a.py":
                return MsgNode({"x": MsgNode(None)})
            else:
                return MsgNode({})
        file_list = [("a.py", "a.py"),
                     ("b.py", "b.py")]
        with patch("trubar.actions.StringCollector.parse_file", parse_file), \
                patch("trubar.actions.walk_files", Mock(return_value=file_list)):
            mess, _ = collect("", {}, "", quiet=True)
            self.assertEqual(mess, dict_to_msg_nodes({"a.py": {"x": None}}))


    # translate: we test walk and StringTranslator; let us assume we call them
    # correctly

    def test_missing_structure(self):
        translations = {
            "a": "b",
            "c": None,
            "d": False,
            "e": True,
            "class `f`": {
                "g": "h",
                "i": False,
                "j": None,
                "k": True},
            "k": "l",
            "def `m`": {"n": "o"},
            "class `p`": {"q": "r"}
        }
        messages = {
            "a": None,
            "c": None,
            "d": None,
            "e": None,
            "class `f`": {
                "g": None,
                "i": None,
                "j": None,
                "k": None},
            "def `k`": {"p": None},
            "m": None,
            "class `p`": {"q": None},
            "s": None,
            "def `t`": {"u": None}
        }
        self.assertEqual(
            yamlized(missing)(translations, messages),
            {"c": None,
             "class `f`": {"j": None},
             "def `k`": {"p": None},
             "m": None,
             "s": None,
             "def `t`": {"u": None}
             }
        )

    def test_missing_pattern(self):
        # test that pattern is applied to first level only
        messages = {
            "foo_1": {"foo_2": "x", "a": {"b": "c"}},
            "2_foo": {"a": {"b": "c"}},
            "bar": {"foo_1": {"b": "c"}}
        }
        self.assertEqual(
            yamlized(missing)({}, messages, "foo"),
            {"foo_1": {"foo_2": "x", "a": {"b": "c"}},
             "2_foo": {"a": {"b": "c"}}}
        )

    def test_merge(self):
        existing = dict_to_msg_nodes({
            "a": None,
            "c": "not-none",
            "d": "x",
            "e": "y",
            "class `f`": {
                "g": "z",
                "i": None,
                "def `j`": {"a": None},
                "k": None},
            "def `k`": {"p": None},
            "m": None,
            "class `p`": {"q": None},
            "s": None,
            "def `t`": {"u": "v"}
        })
        additional = dict_to_msg_nodes({
            "a": "b",
            "c": None,
            "d": False,
            "e": True,
            "class `f`": {
                "g": "h",
                "i": False,
                "j": "r",
                "x": "y",
                "k": True},
            "k": None,
            "def `m`": {"u": "v"},
            "class `p`": {"q": "r"},
            "def `no_trans`": {"x": None, "y": False}
        })
        with io.StringIO() as buf, redirect_stdout(buf):
            removed = merge(additional, existing)
            printed = buf.getvalue()

        self.assertEqual(
            dict_from_msg_nodes(existing),
            {"a": "b",
             "c": "not-none",
             "d": False,
             "e": True,
             "class `f`": {"g": "h",
                 "i": False,
                 "def `j`": {"a": None},
                 "k": True},
             "def `k`": {"p": None},
             "m": None,
             "class `p`": {"q": "r"},
             "s": None,
             "def `t`": {"u": "v"}
            }
        )
        removed = dict_from_msg_nodes(removed)
        self.assertEqual(set(removed), {"class `f`", "def `m`"})
        self.assertEqual(set(removed["class `f`"]), {"x", "j"})
        self.assertEqual(printed,
                         """class `f`/j not in target structure
class `f`/x not in target structure
def `m` not in target structure
""")

        with io.StringIO() as buf, redirect_stdout(buf):
            merge(additional, existing, print_unused=False)
            printed = buf.getvalue()
            self.assertEqual(printed, "")

    def test_template(self):
        messages = {
            "a": "b",
            "c": False,
            "d": True,
            "e": None,
            "f": { "g": "h", "i": False},
            "j": { "k": False, "l": {"m": False, "n": False}}
        }
        self.assertEqual(
            yamlized(template)(messages),
           {"a": None,
            "d": None,
            "e": None,
            "f": { "g": None}
        }
        )
        self.assertEqual(yamlized(template)(messages, "f"), {"f": {"g": None}})
        self.assertEqual(yamlized(template)(messages, "g"), {})

    def test_stat(self):
        messages = {
            "a": "b",
            "c": "not-none",
            "d": False,
            "e": True,
            "class `foo`": {
                "g": "h",
                "i": False,
                "def `j`": {
                    "a": None},
                "k": True},
            "def `k`": {"p": None},
            "m": None,
            "class `p`": {"q": "r"},
            "s": None,
            "def `t`": {"u": "v"}
        }
        stat = yamlized(Stat.collect_stat)(messages)
        self.assertEqual(stat, Stat(5, 2, 4, 2))
        self.assertEqual(abs(stat), 13)

        stat = yamlized(Stat.collect_stat)({})
        self.assertEqual(stat, Stat(0, 0, 0, 0))
        self.assertEqual(abs(stat), 0)


if __name__ == "__main__":
    unittest.main()



import os
import shutil
import tempfile
import unittest

from trubar.messages import dict_to_msg_nodes, dict_from_msg_nodes


def yamlized(func):
    def yamlized_func(*args, **kwargs):
        args = [dict_to_msg_nodes(arg) if isinstance(arg, dict) else arg
                for arg in args]
        res = func(*args, **kwargs)
        return dict_from_msg_nodes(res) if isinstance(res, dict) else res
    return yamlized_func


class TestBase(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.tmpdir = None

    def tearDown(self) -> None:
        super().tearDown()
        if self.tmpdir is not None:
            shutil.rmtree(self.tmpdir)

    def prepare_file(self, filename, s):
        if self.tmpdir is None:
            self.tmpdir = tempfile.mkdtemp()

        fn = os.path.join(self.tmpdir, filename)
        with open(fn, "w", encoding="utf-8") as f:
            f.write(s)
        return fn



import io
from contextlib import redirect_stdout
import unittest
from unittest.mock import patch

from trubar import messages
from trubar.messages import \
    load, dump, dict_to_msg_nodes, dict_from_msg_nodes, MsgNode

from trubar.tests import TestBase, yamlized


class TestUtils(TestBase):
    def test_load_yaml(self):
        fn = self.prepare_file("x.yaml", """
    class `A`:
        foo: bar
        boo: null
        def `f`:
            def `g`:
                vaz: true
                baz: false
                "quoted {dict:03}": "v narekovajih {slovar:04}"
                'quoted {dict:05}': 'v narekovajih {slovar:04}'
    class `B`:
        nothing: nič
    """)
        msgs = load(fn)
        self.assertEqual(
            msgs,
            {'class `A`': MsgNode(value={
                'foo': MsgNode(value='bar'),
                'boo': MsgNode(value=None),
                'def `f`': MsgNode(value={
                    'def `g`': MsgNode(value={
                        'vaz': MsgNode(value=True),
                        'baz': MsgNode(value=False),
                        'quoted {dict:03}': MsgNode(
                            value='v narekovajih {slovar:04}'),
                        'quoted {dict:05}': MsgNode(
                            value='v narekovajih {slovar:04}')
                    })
                })
             }),
             'class `B`': MsgNode(value={
                 'nothing': MsgNode(value='nič')
             })
            })

    def test_load_jaml(self):
        fn = self.prepare_file("x.jaml", """
        class `A`:
            foo: bar: baz
        """)
        msgs = load(fn)
        self.assertEqual(
            msgs,
            {'class `A`': MsgNode(value={
                'foo': MsgNode(value='bar: baz')
            })})

    def test_loader_graceful_exit_on_error(self):
        fn = self.prepare_file("x.jaml", """
        class `A`: asdf
            foo: bar: baz
        """)
        with io.StringIO() as buf, redirect_stdout(buf):
            self.assertRaises(SystemExit, load, fn)
            self.assertIn("rror in", buf.getvalue())
        with io.StringIO() as buf, redirect_stdout(buf):
            self.assertRaises(SystemExit, load, "no such file")
            self.assertIn("not found", buf.getvalue())

    @patch("builtins.print")
    def test_loader_verifies_sanity(self, a_print):
        fn = self.prepare_file("x.jaml", """
        class `A`:
            def `foo`: bar
        """)
        self.assertRaises(SystemExit, load, fn)
        a_print.assert_called()

    def test_check_sanity(self):
        # unexpected namespace
        check_sanity = yamlized(messages.check_sanity)
        with io.StringIO() as buf, redirect_stdout(buf):
            self.assertFalse(check_sanity(
                {"a": "b", "def `f`": {"x": {"y": "z"}}}))
            self.assertEqual(
                buf.getvalue(),
                "def `f`/x: Unexpectedly a namespace\n")
        # def is not a namespace
        with io.StringIO() as buf, redirect_stdout(buf):
            self.assertFalse(check_sanity(
                {"module": {"a": "b", "def `f`": {"def `x`": "y"}}}))
            self.assertEqual(
                buf.getvalue(),
                "module/def `f`/def `x`: Unexpectedly not a namespace\n")
        # class is not a namespace
        with io.StringIO() as buf, redirect_stdout(buf):
            self.assertFalse(check_sanity(
                {"module": {"a": "b", "def `f`": {"class `x`": "y"}}}))
            self.assertEqual(
                buf.getvalue(),
                "module/def `f`/class `x`: Unexpectedly not a namespace\n")
        # everything OK
        with io.StringIO() as buf, redirect_stdout(buf):
            self.assertTrue(check_sanity(
                {"module": {"a": "b", "def `f`": {"class `x`": {"z": "t"}}}}))
            self.assertEqual(
                buf.getvalue(),
                "")
        # check entire structure, even when there are problems
        with io.StringIO() as buf, redirect_stdout(buf):
            self.assertFalse(check_sanity(
                {"module1": {
                    "a": "b",
                    "def `f`": "t",
                    "class `x`": {
                        "z": "t",
                        "m": {"x": False}
                    },
                    "t": "o"},
                 "module2": {
                    "def `g`": None,
                    "x": "y"},
                 "module3": {
                    "a": "b"}
                },
                "somefile"))
            self.assertEqual(
                buf.getvalue(),
                """Errors in somefile:
module1/def `f`: Unexpectedly not a namespace
module1/class `x`/m: Unexpectedly a namespace
module2/def `g`: Unexpectedly not a namespace
""")

    def test_dict_msg_nodes_conversion(self):
        msg_dict = {
            "module1": {
                "a": "b",
                "class `x`": {
                    "z": None,
                    "m": {"x": False}
                },
                "t": True},
            "module2": {
                "x": "y"},
        }
        msg_nodes = {
            'module1': MsgNode(value={
                'a': MsgNode(value='b'),
                    'class `x`': MsgNode(value={
                        'z': MsgNode(value=None),
                        'm': MsgNode(value={
                            'x': MsgNode(value=False)},
                        )}),
                    't': MsgNode(value=True)},
             ),
             'module2': MsgNode(value={
                 'x': MsgNode(value='y')},
            )}
        self.assertEqual(dict_to_msg_nodes(msg_dict), msg_nodes)
        self.assertEqual(dict_from_msg_nodes(msg_nodes), msg_dict)
        self.assertEqual(
            dict_from_msg_nodes({"x": MsgNode("foo", ["bar", "baz"])}),
            {"x": "foo"})

    @patch("builtins.open")
    @patch("yaml.dump")
    @patch("trubar.jaml.dump")
    def test_dump(self, jaml_dump, yaml_dump, _):
        msgdict = {"x": MsgNode("foo", ["bar", "baz"])}
        dump(msgdict, "x.jaml")
        jaml_dump.assert_called_with(msgdict)
        jaml_dump.reset_mock()
        yaml_dump.assert_not_called()

        msgdict = {"x": MsgNode("foo", ["bar", "baz"])}
        dump(msgdict, "x.yaml")
        self.assertEqual(yaml_dump.call_args[0][0], dict_from_msg_nodes(msgdict))
        jaml_dump.assert_not_called()


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from unittest.mock import patch

from trubar.config import config
from trubar.__main__ import check_dir_exists, load_config

import trubar.tests.test_module

test_module_path = os.path.split(trubar.tests.test_module.__file__)[0]


class TestUtils(unittest.TestCase):
    @patch("builtins.print")
    def test_check_dir_exists(self, print_):
        check_dir_exists(test_module_path)
        print_.assert_not_called()

        self.assertRaises(
            SystemExit, check_dir_exists,
            os.path.join(test_module_path, "__init__.py"))
        self.assertIn("not a directory", print_.call_args[0][0])

        self.assertRaises(SystemExit, check_dir_exists, "no_such_path")
        self.assertNotIn("not a directory", print_.call_args[0][0])

    @patch.object(config, "update_from_file")
    def test_load_config(self, update):
        class Args:
            source = os.path.join("x", "y", "z")
            conf = ""
            messages = ""

        args = Args()

        # config is given explicitly
        args.conf = "foo.yaml"
        load_config(args)
        update.assert_called_with("foo.yaml")
        update.reset_mock()
        args.conf = ""

        # .trubarconfig.yaml has precedence over trubar-config.yaml
        with patch("os.path.exists",
                   lambda x: os.path.split(x)[-1] in ("trubar-config.yaml",
                                                      ".trubarconfig.yaml")):
            load_config(args)
        update.assert_called_with(".trubarconfig.yaml")

        # ... but trubar-config.yaml is loaded when there is no trubarconfig.yaml
        with patch("os.path.exists",
                   lambda x: os.path.split(x)[-1] == "trubar-config.yaml"):
            load_config(args)
        update.assert_called_with("trubar-config.yaml")

        # Load from source directory
        confile = os.path.join(args.source, ".trubarconfig.yaml")
        with patch("os.path.exists", lambda x: x == confile):
            load_config(args)
        update.assert_called_with(confile)

        # Messages directory has precedence over source
        mess_dir = os.path.join("t", "u", "v")
        args.messages = os.path.join(mess_dir, "mess.yaml")
        confile = os.path.join(mess_dir, ".trubarconfig.yaml")
        with patch("os.path.exists", lambda x: x == confile):
            load_config(args)
        update.assert_called_with(confile)


if __name__ == "__main__":
    unittest.main()



import re


class JamlError(Exception):
    pass


def readfile(name, encoding="utf8"):
    with open(name, encoding=encoding) as f:
        return read(f.read())


def read(text):
    return readlines(text.splitlines())


def readlines(lines):
    # prevent circular import, pylint: disable=import-outside-toplevel
    from trubar.messages import MsgNode

    def error(msg, line=None):
        raise JamlError(f"Line {line or lineno}: {msg}") from None

    def read_quoted(line):
        nonlocal lineno
        start_line = lineno
        block = ""
        q, line = line[0], line[1:]
        while (mo := re.match(f"((?:[^{q}]|(?:{q}{q}))*){q}(?!{q})(.*)", line)
                ) is None:
            block += line + "\n"
            try:
                lineno, line = next(linegen)
            except StopIteration:
                error("file ends before the end of quoted string", start_line)
        inside, after = mo.groups()
        block += inside
        return block.replace(2 * q, q), after

    def check_no_comments():
        if comments:
            error("stray comment", comment_start)

    items = {}
    stack = [(-1, items, True)]
    comments = []
    comment_start = None
    linegen = enumerate(lines, start=1)
    lineno, line = 0, ""  # for error reporting for empty files
    for lineno, line in linegen:
        # Skip empty lines
        if not line.strip():
            continue
        sline = line.lstrip()
        indent = len(line) - len(sline)
        line = sline

        # Indentation
        last_indent, _, indent_expected = stack[-1]
        if indent_expected:
            if indent <= last_indent:
                error("indent expected")
            stack.append((indent, stack.pop()[1], False))
        elif indent > last_indent:
            error("unexpected indent")
        elif indent < last_indent:
            check_no_comments()
            while stack and indent != stack[-1][0]:
                stack.pop()
            if not stack:
                error("unindent does not match any outer level")

        # Gather comments
        if line[0] == "#":
            comments.append(line)
            comment_start = comment_start or lineno
            continue

        # Get key
        if line[0] in "'\"":
            key, after = read_quoted(line)
            if after[:2] != ": ":
                error("quoted key must be followed by a ': '")
            else:
                value = after[2:].lstrip()
        else:
            mo = re.match(r"(.*?):(?:\s+|$)(.*)", line)
            if mo is None:
                if ":" in line:
                    raise error("colon at the end of the key should be "
                                "followed by a space or a new line")
                else:
                    raise error("key followed by colon expected")
            key, value = mo.groups()

        # Get value
        # `value` is lstripped, but may contain whitespace at the end, which is
        # included in quoted values
        # Leaves
        if value.strip():
            if value[0] in "'\"":
                value, after = read_quoted(value)
                if after.strip():
                    error("quoted value must be followed by end of line")
            else:
                value = {"true": True, "false": False, "null": None
                         }.get(value.strip(), value)
            stack[-1][1][key] = MsgNode(value, comments or None)
        # Internal nodes
        else:
            space = {}
            stack[-1][1][key] = MsgNode(space, comments or None)
            stack.append((indent, space, True))

        comments = []
        comment_start = None

    if stack[-1][-1]:
        raise error("unexpected end of file")

    check_no_comments()
    return items


def dump(d, indent=""):
    def quotescape(s, allow_colon):
        if not s \
                or "\n" in s \
                or ": " in s and not allow_colon \
                or s[0] in " #\"'|" \
                or s[-1] in " \t\n":
            q = '"' if "'" in s else "'"
            return f"{q}{s.replace(q, 2 * q)}{q}"
        return s

    def dumpval(s):
        trans = {True: "true", False: "false", None: "null", "": '""'}
        if s in trans:
            return trans[s]
        return quotescape(s, True)

    res = ""
    for key, node in d.items():
        if node.comments is not None:
            res += "".join(f"{indent}{comment}\n" for comment in node.comments)
        if isinstance(node.value, dict):
            res += f"{indent}{key}:\n{dump(node.value, indent + '    ')}"
        else:
            res += f"{indent}{quotescape(key, False)}: {dumpval(node.value)}\n"
    return res



import argparse
import os
import sys

from trubar.actions import \
    collect, translate, merge, missing, template, stat, \
    ReportCritical
from trubar.messages import load, dump
from trubar.config import config
from trubar.utils import check_any_files, dump_removed


def check_dir_exists(path):
    if not os.path.isdir(path):
        if os.path.exists(path):
            print(f"{path} is not a directory.")
        else:
            print(f"Directory {path} does not exist.")
        sys.exit(2)


def load_config(args):
    if args.conf:
        config.update_from_file(args.conf)
        return

    paths = [""]
    if getattr(args, "messages", False):
        paths.append(os.path.split(args.messages)[0])
    if getattr(args, "source", False):
        paths.append(args.source)
    for path in paths:
        for name in (".trubarconfig.yaml", "trubar-config.yaml"):
            fullname = os.path.join(path, name)
            if os.path.exists(fullname):
                config.update_from_file(fullname)
                return


def main() -> None:
    def add_parser(name, desc):
        subparser = subparsers.add_parser(name, help=desc, description=desc)
        subparser.add_argument(
            "-p", "--pattern", default="", metavar="pattern",
            help="include only files whose full path include the pattern")
        return subparser

    argparser = argparse.ArgumentParser()
    argparser.add_argument(
        "--conf", default="", metavar="configuration-file",
        help="configuration file")
    subparsers = argparser.add_subparsers(required=True, dest="action")

    parser = add_parser("collect", "Collect message strings in source files")
    parser.add_argument(
        "-s", "--source", metavar="source-dir",
        required=True, help="source path")
    parser.add_argument(
        "-u", "--newer", action="store_true",
        help="check only source files that are newer than the message file")
    parser.add_argument(
        "messages", metavar="messages",
        help="existing or new file with messages")
    parser.add_argument(
        "-r", "--removed", metavar="removed-translations", default=None,
        help="file for removed translations, if any")
    parser.add_argument(
        "-q", "--quiet", action="store_true",
        help="supress intermediary outputs")
    parser.add_argument(
        "-n", "--dry-run", action="store_true",
        help="don't write the output file; removed translations (if any) are written"
    )

    parser = add_parser("translate", "Prepare sources with translations")
    parser.add_argument(
        "messages", metavar="messages",
        help="file with translated messages")
    parser.add_argument(
        "-d", "--dest", metavar="destination-dir",
        help="destination path")
    parser.add_argument(
        "-i", "--inplace", action="store_true",
        help="translate files in-place")
    parser.add_argument(
        "-s", "--source", metavar="source-dir",
        required=True, help="source path")
    parser.add_argument(
        "--static", metavar="static-files-dir", action="append",
        help="directory(-ies) with static files to copy")
    parser.add_argument(
        "-q", "--quiet", action="store_true",
        help="supress intermediary outputs")
    parser.add_argument(
        "-v", "--verbosity", type=int, choices=range(4), default=1,
        help="verbosity (0=quiet, 1=updates, 2=translations, 3=all")
    parser.add_argument(
        "-n", "--dry-run", action="store_true",
        help="don't write anything; perform a trial run to check the structure"
    )

    parser = add_parser("merge",
                        "Merge translations into template or existing "
                        "translations")
    parser.add_argument(
        "translations", metavar="translations",
        help="new or updated translations")
    parser.add_argument(
        "messages", metavar="messages",
        help="file with messages; this file is updated unless --output is given")
    parser.add_argument(
        "-o", "--output", metavar="output-file",
        help="output file; if omitted, existing file will updated")
    parser.add_argument(
        "-u", "--unused", metavar="unused", default=None,
        help="file for unused translations (if any)")
    parser.add_argument(
        "-n", "--dry-run", action="store_true",
        help="run without writing anything"
    )

    parser = add_parser("template",
                        "Create empty template from existing translations")
    parser.add_argument(
        "messages", metavar="translations",
        help="file with existing translations for another language")
    parser.add_argument(
        "-o", "--output", metavar="output-file", required=True,
        help="output file")

    parser = add_parser("missing", "Prepare a file with missing translations")
    parser.add_argument(
        "messages", metavar="messages",
        help="file with existing translations")
    parser.add_argument(
        "-m", "--all-messages", metavar="all-messages", required=False,
        help="all messages")
    parser.add_argument(
        "-o", "--output", metavar="output-file", required=True,
        help="missing translations")

    parser = add_parser("stat", "Show statistics about messages in the file")
    parser.add_argument(
        "messages", metavar="messages",
        help="file with messages")

    args = argparser.parse_args(sys.argv[1:])
    load_config(args)
    pattern = args.pattern

    if args.action == "collect":
        min_time = None
        check_dir_exists(args.source)
        if os.path.exists(args.messages):
            existing = load(args.messages)
            check_any_files(existing, args.source)
            if args.newer:
                min_time = os.stat(args.messages).st_mtime
        else:
            existing = {}
        messages, removed = collect(args.source, existing, pattern,
                                    quiet=args.quiet, min_time=min_time)
        if not args.dry_run:
            dump(messages, args.messages)
        dump_removed(removed, args.removed, args.messages)

    elif args.action == "translate":
        check_dir_exists(args.source)
        if args.inplace:
            if args.dest:
                argparser.error("options -d and -i are incompatible")
            else:
                args.dest = args.source
        elif not args.dest:
            argparser.error("specify destination (-d) or translate in place (-i)")
        if args.static:
            config.set_static_files(args.static)
        verbosity = ReportCritical if args.quiet else args.verbosity
        messages = load(args.messages)
        check_any_files(messages, args.source)
        translate(messages, args.source, args.dest, pattern,
                  verbosity=verbosity, dry_run=args.dry_run)

    elif args.action == "merge":
        additional = load(args.translations)
        existing = load(args.messages)
        unused = merge(additional, existing, pattern,
                       print_unused=not args.unused)
        if not args.dry_run:
            dump(existing, args.output or args.messages)
        if args.unused and unused:
            dump(unused, args.unused)

    elif args.action == "template":
        existing = load(args.messages)
        new = template(existing, pattern)
        dump(new, args.output)

    elif args.action == "missing":
        translations = load(args.messages)
        messages = load(args.all_messages) \
            if args.all_messages else translations
        needed = missing(translations, messages, pattern)
        dump(needed, args.output)

    elif args.action == "stat":
        messages = load(args.messages)
        stat(messages, pattern)


if __name__ == "__main__":
    main()



import unittest
from unittest.mock import patch

from trubar import jaml
from trubar.messages import MsgNode


class JamlReaderTest(unittest.TestCase):
    def test_read(self):
        text = """
# jaml file
class `A`:
    foo: bar
    boo: null
    # This function contains nothing but another function
    def `f`:
    
        def `g`:
            # I think that vaz is just
            # mistyped baz
            
            vaz: true
            baz: false
            "quoted {dict:03}": "v narekovajih {slovar:04}"
            'quoted {dict:05}': 'v narekovajih {slovar:04}'
            'quoted {dict:06}': v narekovajih {slovar:04}
            "This is so: so": to je tako: tako
# Yet another class
class `B`:
    nothing: nič
"""
        msgs = jaml.read(text)
        self.assertEqual(
            msgs,
            {'class `A`': MsgNode(comments=["# jaml file"], value={
                'foo': MsgNode(value='bar'),
                'boo': MsgNode(value=None),
                'def `f`': MsgNode(
                    comments=['# This function contains nothing '
                              'but another function'],
                    value={
                        'def `g`': MsgNode(value={
                            'vaz': MsgNode(
                                comments=['# I think that vaz is just',
                                          '# mistyped baz'],
                                value=True),
                            'baz': MsgNode(value=False),
                            'quoted {dict:03}': MsgNode(
                                value='v narekovajih {slovar:04}'),
                            'quoted {dict:05}': MsgNode(
                                value='v narekovajih {slovar:04}'),
                            'quoted {dict:06}': MsgNode(
                                value='v narekovajih {slovar:04}'),
                            'This is so: so': MsgNode(
                                value='to je tako: tako')
                        })
                    }
                )
             }),
             'class `B`': MsgNode(
                 comments=['# Yet another class'],
                 value={'nothing': MsgNode(value='nič')
             })
            })

    def test_read_empty_file(self):
        self.assertRaisesRegex(jaml.JamlError, "unexpected end", jaml.read, "")

    def test_read_quoted_blocks(self):
        self.assertEqual(jaml.read('''a/b.py:
    def `f`:
        "a
b
c": abc
        abc: "
     a
''' + " " * 5 + '''
   bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb
"
        foo: false
        def: "a
bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"
x/y.py: null
        '''),
         {"a/b.py":
             MsgNode(
                 value={"def `f`": MsgNode({
                     "a\nb\nc": MsgNode("abc"),
                     "abc": MsgNode(
                         "\n     a\n     \n   " + "b" * 100 + "\n"),
                     "foo": MsgNode(False),
                     "def": MsgNode("a\n" + "b" * 100),
                 })}),
             "x/y.py": MsgNode(None)
         }
    )

    def test_read_quotes_in_values(self):
        text = '''
foo1: "bar"
foo2: """bar"
foo4: "ba""r"
foo5: "bar"""
foo7: 'bar' 
foo8: 'ba''r' 
foo9: 'ba"r'
foo12: "bar''" 
'''
        msgs = jaml.read(text)
        self.assertEqual(
            [node.value for node in msgs.values()],
            ["bar", "\"bar", "ba\"r", "bar\"", "bar", "ba'r", 'ba"r', "bar''"])

    def test_read_quotes_in_keys(self):
        text = """
1bar": foo2
"2bar": foo3
"3ba""r": foo4
4bar': foo6
'5bar': foo7
'6ba''r': foo8
"7ba""r": foo12
'8ba''r': foo13
"""
        msgs = jaml.read(text)
        self.assertEqual(
            list(msgs),
            ['1bar"', '2bar', '3ba"r', "4bar'", "5bar", "6ba'r", '7ba"r', "8ba'r"])

    def test_read_colons_in_keys(self):
        text = """
bar:baz: foo1
bar: baz: foo2
"bar:baz: boz": foo3
"bar"": baz: boz": foo4
        """
        msgs = jaml.read(text)
        self.assertEqual(
            list(msgs),
            ["bar:baz", "bar", "bar:baz: boz", "bar\": baz: boz"]
        )

    def test_stray_comments(self):
        self.assertRaisesRegex(jaml.JamlError, "5", jaml.read, """
class `A`:
    def `f`:
        sth: sth
        # stray comment
    def `g`:
        sth:stg
        """)
        self.assertRaisesRegex(jaml.JamlError, "4", jaml.read, """
class `A`:
    def `f`:
        # stray comment""")
        self.assertRaisesRegex(jaml.JamlError, "4", jaml.read, """
class `A`:
    def `f`:
        # stray comment
""")
        self.assertRaisesRegex(jaml.JamlError, "4", jaml.read, """
class `A`:
    def `f`:
        # stray comment
                
                
""")

    def test_indentation_errors(self):
        # This function checks for exact error messages. While this is not
        # a good practice in general, it makes these tests more readable
        self.assertRaisesRegex(
            jaml.JamlError, "Line 4: unexpected indent", jaml.read, """
        abc:
            def: fgh
                ghi: jkl
            jkl: mno
        prs:
            tuv: bdf""",
        )
        self.assertRaisesRegex(
            jaml.JamlError, "Line 4: unindent does not match any outer level",
            jaml.read, """
        abc:
            def: fgh
          ghi: jkl
            jkl: mno
        prs:
            tuv: bdf""",
        )
        self.assertRaisesRegex(
            jaml.JamlError, "Line 4: unindent does not match any outer level",
            jaml.read, """
        abc:
            def: fgh
ghi: jkl
            jkl: mno
        prs:
            tuv: bdf""",
        )
        self.assertRaisesRegex(
            jaml.JamlError, "Line 4: unindent does not match any outer level",
            jaml.read, """
        abc:
            def: fgh
ghi: jkl
            jkl: mno
        prs:
            tuv:""",
        )
        self.assertRaisesRegex(
            jaml.JamlError, "Line 9: unexpected end of file", jaml.read, """
        abc:
            def: fgh
            ghi: jkl
            jkl: mno
        prs:
            tuv:
            
            """,
        )

    def test_syntax_errors(self):
        self.assertRaisesRegex(
            jaml.JamlError, "Line 1: file ends.*", jaml.read, "'''x: y")
        self.assertRaisesRegex(
            jaml.JamlError, "Line 1: file ends.*", jaml.read, "'x: y")
        self.assertRaisesRegex(
            jaml.JamlError, "Line 2: quoted key must be followed .*",
            jaml.read, '"x\ny"\na:b')
        self.assertRaisesRegex(
            jaml.JamlError, "Line 2: quoted value must be followed .*",
            jaml.read, 'x: "\na": b')
        self.assertRaisesRegex(
            jaml.JamlError,
            "Line 1: colon at the end of the key should be "
            "followed by a space or a new line", jaml.read, "x:y")

    def test_format_errors(self):
        # This function checks for exact error messages. While this is not
        # a good practice in general, it makes these tests more readable
        self.assertRaisesRegex(
            jaml.JamlError, "Line 3: key followed by colon expected", jaml.read, """
        abc:
            def
                ghi: jkl
            jkl: mno
        prs:
            tuv: bdf""",
        )
        self.assertRaisesRegex(
            jaml.JamlError, "Line 4: file ends", jaml.read, """
        abc:
            def:
                "ghi: jkl
            jkl: mno
        prs:
            tuv: bdf""",
        )

    @patch("trubar.jaml.read")
    def test_readfile(self, read):
        jaml.readfile(jaml.__file__)
        with open(jaml.__file__, encoding="utf-8") as f:
            read.assert_called_with(f.read())


class JamlDumperTest(unittest.TestCase):
    def test_dump(self):
        tree = {"a/b.py":
                    MsgNode(comments=["# a few", "# initial comments"],
                       value={ "def `f`":
                                   MsgNode(comments=["# a function!"],
                                           value={"foo": MsgNode("bar"),
                                                  "baz": MsgNode(None, ["# eh"])}),
                               "yada": MsgNode(comments=["# bada", "# boom"],
                                               value=True),
                               "": MsgNode(""),
                               }),
                "class `A`":
                    MsgNode(value=False)}

        self.assertEqual(
            jaml.dump(tree),
            """
# a few
# initial comments
a/b.py:
    # a function!
    def `f`:
        foo: bar
        # eh
        baz: null
    # bada
    # boom
    yada: true
    '': ""
class `A`: false
"""[1:])

    def test_backslashes(self):
        self.assertEqual(jaml.dump({r"a\nb": MsgNode(r"c\nd")}).strip(),
                         r"a\nb: c\nd")

    def test_dump_quotes(self):
        self.assertEqual(jaml.dump({"'foo'": MsgNode("'asdf'")}),
                         """"'foo'": "'asdf'"\n""")
        self.assertEqual(jaml.dump({'"foo"': MsgNode('"asdf"')}),
                         """'"foo"': '"asdf"'\n""")
        self.assertEqual(jaml.dump({"'foo": MsgNode("asdf'")}),
                         """\"'foo": asdf'\n""")

    def test_dump_spaces_in_value(self):
        self.assertEqual(jaml.dump({"foo": MsgNode("bar ")}),
                         "foo: 'bar '\n")
        self.assertEqual(jaml.dump({"foo": MsgNode(" bar")}),
                         "foo: ' bar'\n")

    def test_quoting_keys(self):
        self.assertEqual(jaml.dump({"| ": MsgNode(True)}),
                         "'| ': true\n")
        self.assertEqual(jaml.dump({"# ": MsgNode(True)}),
                         "'# ': true\n")
        self.assertEqual(jaml.dump({" x": MsgNode(True)}),
                         "' x': true\n")
        self.assertEqual(jaml.dump({"x ": MsgNode(True)}),
                         "'x ': true\n")
        self.assertEqual(jaml.dump({" x: y": MsgNode(True)}),
                         "' x: y': true\n")
        self.assertEqual(jaml.dump({"x:y": MsgNode(True)}),
                         "x:y: true\n")

    def test_dump_blocks(self):
        tree = {"a/b.py":
                    MsgNode(
                       value={ "def `f`": MsgNode({
                           "a\nb\nc": MsgNode("abc"),
                           "abc": MsgNode("\n     a\n     \n   " + "b" * 100 + "\n"),
                           "foo": MsgNode(False),
                           "def": MsgNode("a\n" + "b" * 100),
                       })}),
                "x/y.py": MsgNode(None)
                }
        self.assertEqual(jaml.dump(tree), '''a/b.py:
    def `f`:
        'a
b
c': abc
        abc: '
     a
''' + " " * 5 + '''
   bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb
'
        foo: false
        def: 'a
bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb'
x/y.py: null
''')


if __name__ == "__main__":
    unittest.main()



import os
from pathlib import PureWindowsPath

import unittest

from unittest.mock import patch, Mock

from trubar.utils import walk_files, check_any_files, unique_name, dump_removed
from trubar.config import config
import trubar.tests.test_module


test_module_path = os.path.split(trubar.tests.test_module.__file__)[0]


class UtilsTest(unittest.TestCase):
    def test_walk_files(self):
        tmp = test_module_path
        old_pattern = config.exclude_pattern
        try:
            config.set_exclude_pattern("")
            self.assertEqual(
                set(walk_files(tmp, select=True)),
                {('bar_module/__init__.py',
                  f'{tmp}/bar_module/__init__.py'),
                 ('bar_module/foo_module/__init__.py',
                  f'{tmp}/bar_module/foo_module/__init__.py'),
                 ('baz_module/__init__.py',
                  f'{tmp}/baz_module/__init__.py'),
                 ('__init__.py',
                  f'{tmp}/__init__.py')}
            )

            old_path = os.getcwd()
            try:
                os.chdir(tmp)
                self.assertEqual(
                    set(walk_files(".", select=True)),
                    {('bar_module/__init__.py',
                      './bar_module/__init__.py'),
                     ('bar_module/foo_module/__init__.py',
                      './bar_module/foo_module/__init__.py'),
                     ('baz_module/__init__.py',
                      './baz_module/__init__.py'),
                     ('__init__.py',
                      './__init__.py')}
                )
                self.assertEqual(
                    set(walk_files(".", "bar", select=True)),
                    {('bar_module/__init__.py',
                      './bar_module/__init__.py'),
                     ('bar_module/foo_module/__init__.py',
                      './bar_module/foo_module/__init__.py')}
                )
            finally:
                os.chdir(old_path)

            self.assertEqual(
                set(walk_files(tmp, select=False)),
                {('bar_module/__init__.py',
                  f'{tmp}/bar_module/__init__.py'),
                 ('bar_module/foo_module/__init__.py',
                  f'{tmp}/bar_module/foo_module/__init__.py'),
                 ('baz_module/__init__.py',
                  f'{tmp}/baz_module/__init__.py'),
                 ('__init__.py',
                  f'{tmp}/__init__.py'),
                 ('baz_module/not-python.js',
                  f'{tmp}/baz_module/not-python.js')}
            )

            config.set_exclude_pattern("b?r_")
            self.assertEqual(
                set(walk_files(tmp, select=False)),
                {('bar_module/__init__.py',
                  f'{tmp}/bar_module/__init__.py'),
                 ('bar_module/foo_module/__init__.py',
                  f'{tmp}/bar_module/foo_module/__init__.py'),
                 ('baz_module/__init__.py',
                  f'{tmp}/baz_module/__init__.py'),
                 ('__init__.py',
                  f'{tmp}/__init__.py'),
                 ('baz_module/not-python.js',
                  f'{tmp}/baz_module/not-python.js')}
            )
            self.assertEqual(
                set(walk_files(tmp, select=True)),
                {('baz_module/__init__.py',
                  f'{tmp}/baz_module/__init__.py'),
                 ('__init__.py',
                  f'{tmp}/__init__.py')}
            )
        finally:
            config.set_exclude_pattern(old_pattern)

    @patch("os.walk",
           return_value=[(r"c:\foo\bar\ann\bert",
                          None,
                          ["cecil.py", "dan.txt", "emily.py"])]
           )
    @patch("os.path.join", new=lambda *c: "\\".join(c))
    @patch("trubar.utils.PurePath", new=PureWindowsPath)
    def test_walk_backslashes_on_windows(self, _):
        self.assertEqual(
            list(walk_files(r"c:\foo\bar", select=False)),
            [("ann/bert/cecil.py", r"c:\foo\bar\ann\bert\cecil.py"),
             ("ann/bert/dan.txt", r"c:\foo\bar\ann\bert\dan.txt"),
             ("ann/bert/emily.py", r"c:\foo\bar\ann\bert\emily.py")]
        )
        self.assertEqual(
            list(walk_files(r"c:\foo\bar", "l", select=False)),
            [("ann/bert/cecil.py", r"c:\foo\bar\ann\bert\cecil.py"),
             ("ann/bert/emily.py", r"c:\foo\bar\ann\bert\emily.py")])
        # Don't match pattern in path
        self.assertEqual(
            list(walk_files(r"c:\foo\bar", "o", select=False)),
            [])

    @patch("sys.exit")
    @patch("builtins.print")
    def test_check_any_files(self, print_, exit_):
        keys = "a/x.py a/y.py a/b/x.py a/b/z.py".split()
        translations = dict.fromkeys(keys)
        with patch("trubar.utils.walk_files",
                   Mock(return_value=[(k, k) for k in keys])):
            check_any_files(translations, "foo/bar")
            exit_.assert_not_called()
            print_.assert_not_called()

        with patch("trubar.utils.walk_files",
                   Mock(return_value=[("t/x/" + k, k) for k in keys])):
            check_any_files(translations, "foo/bar")
            exit_.assert_called()
            print_.assert_called()
            msg = print_.call_args[0][0]
            self.assertIn("-s foo/bar/t/x", msg)
            print_.reset_mock()
            exit_.reset_mock()

            home = os.path.expanduser("~/foo/bar")
            check_any_files(translations, home)
            exit_.assert_called()
            msg = print_.call_args[0][0]
            self.assertIn("-s ~/foo/bar/t/x", msg)
            print_.reset_mock()
            exit_.reset_mock()

            with patch("sys.platform", "win32"):
                check_any_files(translations, home)
                exit_.assert_called()
                print_.assert_called()
                msg = print_.call_args[0][0]
                print_.assert_called()
                self.assertIn(f"-s {home}/t/x", msg)
                print_.reset_mock()
                exit_.reset_mock()

        keys = "a/x.py a/y.py a/b/x.py a/b/z.py a/b/u.py".split()
        with patch("trubar.utils.walk_files",
                   Mock(return_value=[(k, k) for k in keys])):
            check_any_files(dict.fromkeys(["x.py", "z.py", "u.py"]), "foo")
            exit_.assert_called()
            print_.assert_called()
            msg = print_.call_args[0][0]
            self.assertIn("-s foo/a/b", msg)
            print_.reset_mock()
            exit_.reset_mock()

        keys = "a/x.py a/y.py a/b/x.py a/b/z.py".split()
        with patch("trubar.utils.walk_files",
                   Mock(return_value=[(k, k) for k in keys])):
            check_any_files(dict.fromkeys(["x.py", "z.py", "u.py"]), "foo")
            exit_.assert_called()
            print_.assert_called()
            msg = print_.call_args[0][0]
            self.assertNotIn("-s foo/a/b", msg)
            print_.reset_mock()
            exit_.reset_mock()

    def test_unique_name(self):
        with patch("os.path.exists", return_value=False):
            self.assertEqual(unique_name("some name.yaml"),
                             "some name.yaml")

        with patch("os.path.exists", return_value=True):
            with patch("os.listdir", return_value=["some name.yaml"]) as listdir:
                self.assertEqual(unique_name("some name.yaml"),
                                 "some name (1).yaml")
                listdir.assert_called_with(".")

                self.assertEqual(unique_name("abc/def/some name.yaml"),
                                 "abc/def/some name (1).yaml")
                listdir.assert_called_with("abc/def")

            with patch("os.listdir", return_value=["some name.yaml",
                                                   "some name (1).yaml",
                                                   "some name (2).yaml",
                                                   "non sequitur.yaml",
                                                   ]) as listdir:
                self.assertEqual(unique_name("some name.yaml"),
                                 "some name (3).yaml")
                listdir.assert_called_with(".")

            with patch("os.listdir", return_value=["some name.yaml",
                                                   "some name (1).yaml",
                                                   "non sequitur.yaml",
                                                   "some name (4).yaml",
                                                   ]) as listdir:
                self.assertEqual(unique_name("some name.yaml"),
                                 "some name (5).yaml")
                listdir.assert_called_with(".")

    @patch("trubar.utils.dump")
    def test_dump_removed(self, mock_dump):
        dump_removed({}, "removed.yaml", "abc/def/x.yaml")
        mock_dump.assert_not_called()

        msgs = Mock()
        dump_removed(msgs, "removed.yaml", "abc/def/x.yaml")
        mock_dump.assert_called_with(msgs, "removed.yaml")

        dump_removed(msgs, None, "abc/def/xyz.jaml")
        mock_dump.assert_called_with(msgs, "abc/def/removed-from-xyz.jaml")


if __name__ == "__main__":
    unittest.main()



import re

import os
import sys
from pathlib import PurePath
from typing import Iterator, Tuple, Optional

from trubar.config import config
from trubar.messages import MsgDict, dump


def walk_files(path: str, pattern: str = "", *, select: bool
               ) -> Iterator[Tuple[str, str]]:
    path = os.path.normpath(path)
    for dirpath, _, files in sorted(os.walk(path)):
        for name in sorted(files):
            if select and not name.endswith(".py"):
                continue
            name = os.path.join(dirpath, name)
            keyname = PurePath(name[len(path) + 1:]).as_posix()
            if pattern in keyname and \
                    not (select and config.exclude_re
                         and config.exclude_re.search(keyname)):
                yield keyname, name


def check_any_files(translations: MsgDict, path: str):
    source_keys = {n for n, _ in walk_files(path, "", select=True)}
    trans_keys = set(translations)
    if not trans_keys or source_keys & trans_keys:
        return
    suggestion = ""
    best_matched = 2  # Require at least three matches to make a suggestion
    tried = set()
    for fname in source_keys:
        start = ""
        for part in fname.split("/")[:3]:
            start += part + "/"
            if start in tried:
                continue
            tried.add(start)
            matched = len(source_keys & {start + k for k in translations})
            if matched > best_matched:
                best_matched = matched
                suggestion = start
    if suggestion:
        suggestion = os.path.join(path, suggestion[:-1])
        if sys.platform != "win32":
            home = os.path.expanduser("~")
            if suggestion.startswith(home):
                suggestion = "~/" + suggestion[len(home) + 1:]
        suggestion = f"; try -s {suggestion} instead"
    print("Paths in translations do not match any existing files.\n"
          f"One reason may be an incorrect source path{suggestion}.")
    sys.exit(5)


def unique_name(name: str) -> str:
    if not os.path.exists(name):
        return name

    path, name = os.path.split(name)
    base, ext = os.path.splitext(name)
    pattern = fr"{re.escape(base)} \((\d+)\){ext}"
    existing = (re.match(pattern, fname) for fname in os.listdir(path or "."))
    ver = 1 + max((int(mo.group(1)) for mo in existing if mo), default=0)
    return os.path.join(path, f"{base} ({ver}){ext}")


def dump_removed(removed: MsgDict,
                 removed_name: Optional[str], name: str) -> None:
    if not removed:
        return
    if not removed_name:
        path, name = os.path.split(name)
        removed_name = os.path.join(path, "removed-from-" + name)
    removed_name = unique_name(removed_name)
    dump(removed, removed_name)



import dataclasses
import unittest
from unittest.mock import patch

from trubar.config import Configuration
from trubar.tests import TestBase


class ConfigTest(TestBase):
    def prepare(self, s):
        # pylint: disable=attribute-defined-outside-init
        self.fn = self.prepare_file("test.yaml", s)

    def test_proper_file(self):
        config = Configuration()
        self.prepare("smart_quotes: false\n\nencoding: cp-1234")
        config.update_from_file(self.fn)
        self.assertFalse(config.smart_quotes)
        self.assertTrue(config.auto_prefix)
        self.assertEqual(config.encoding, "cp-1234")

    @patch("builtins.print")
    def test_malformed_file(self, _):
        config = Configuration()
        self.prepare("smart_quotes: false\n\nencoding")
        self.assertRaises(SystemExit, config.update_from_file, self.fn)

    @patch("builtins.print")
    def test_unrecognized_option(self, a_print):
        config = Configuration()
        self.prepare("smart_quotes: false\n\nauto_magog: false")
        self.assertRaises(SystemExit, config.update_from_file, self.fn)
        self.assertIn("auto_magog", a_print.call_args[0][0])

    @patch("builtins.print")
    def test_invalid_type(self, a_print):
        # At the time of writing, Configuration had only bool and str settings,
        # which can never fail on conversion. To reach that code in test, we
        # imagine setting that can
        # Data classes do not support inheritance, so we patch the tested method
        # into a new class.
        @dataclasses.dataclass
        class ConfigurationWithInt:
            encoding: str = "ascii"
            foo: int = 42
            update_from_file = Configuration.update_from_file

        config = ConfigurationWithInt()
        self.prepare("foo: bar")
        self.assertRaises(SystemExit, config.update_from_file, self.fn)
        self.assertIn("foo", a_print.call_args[0][0])

    @patch("builtins.print")
    def test_static_files(self, _):
        config = Configuration()
        self.assertEqual(len(config.static_files), 0)

        self.prepare("static-files: static_files_lan")
        with patch("os.path.exists", lambda x: "test.yaml" in x):
            self.assertRaises(SystemExit, config.update_from_file, self.fn)

        with patch("os.path.exists", lambda _: True):
            config.update_from_file(self.fn)
            self.assertEqual(config.static_files, ("static_files_lan", ))

        self.prepare("static-files:\n"
                     "- static_files_lan\n"
                     "- ban\n"
                     "- pet_podgan\n")
        with patch("os.path.exists", lambda x: "test.yaml" in x or "lan" in x):
            self.assertRaises(SystemExit, config.update_from_file, self.fn)

        with patch("os.path.exists", lambda x: "test.yaml" in x or "an" in x):
            config.update_from_file(self.fn)
            self.assertEqual(config.static_files,
                             ("static_files_lan", "ban", "pet_podgan"))

    def test_exclude(self):
        config = Configuration()
        self.assertTrue(config.exclude_re.search("dir/tests/test_something.py"))
        self.prepare("exclude-pattern: ba?_m")
        config.update_from_file(self.fn)
        self.assertFalse(config.exclude_re.search("dir/tests/test_something.py"))
        self.prepare("exclude-pattern: ")
        config.update_from_file(self.fn)
        self.assertIsNone(config.exclude_re)

        config.set_exclude_pattern("tests/test_")
        self.assertTrue(config.exclude_re.search("dir/tests/test_something.py"))
        config.set_exclude_pattern("")
        self.assertIsNone(config.exclude_re)


if __name__ == "__main__":
    unittest.main()



import os
import sys
import re
from typing import NamedTuple, Union, Optional, Dict, List

import yaml

from trubar import jaml
from trubar.config import config


class MsgNode(NamedTuple):
    value: Union["MsgDict", str, bool, None]
    comments: Optional[List[str]] = None


MsgDict = Dict[str, MsgNode]
PureDict = Dict[str, Union[bool, None, str, "PureDict"]]


def load(filename: str) -> MsgDict:
    if not os.path.exists(filename):
        print(f"File not found: {filename}")
        sys.exit(2)
    is_yaml = os.path.splitext(filename)[1] == ".yaml"
    try:
        if is_yaml:
            with open(filename, encoding=config.encoding) as f:
                messages = yaml.load(f, Loader=yaml.Loader)
        else:
            messages = jaml.readfile(filename, encoding=config.encoding)
    except (jaml.JamlError, yaml.YAMLError) as exc:
        print(f"Error in {filename}:\n{exc}")
        sys.exit(3)
    if is_yaml:
        messages = dict_to_msg_nodes(messages)
    if not check_sanity(messages, filename):
        sys.exit(4)
    return messages


def check_sanity(message_dict: MsgDict, filename: Optional[str] = None):
    key_re = re.compile(r"^((def)|(class)) `\w+`")
    sane = True

    def fail(msg):
        nonlocal sane
        if sane and filename is not None:
            print(f"Errors in {filename}:")
        print(msg)
        sane = False

    def check_sane(messages: MsgDict, path: str):
        for key, obj in messages.items():
            npath = f"{path}/{key}"
            if key_re.fullmatch(key) is None:
                if isinstance(obj.value, dict):
                    fail(f"{npath}: Unexpectedly a namespace")
            else:
                if not isinstance(obj.value, dict):
                    fail(f"{npath}: Unexpectedly not a namespace")
                else:
                    check_sane(obj.value, f"{npath}")

    for fname, fspace in message_dict.items():
        if isinstance(fspace.value, dict):
            check_sane(fspace.value, fname)
    return sane


def dict_to_msg_nodes(messages: PureDict) -> Dict[str, MsgDict]:
    return {
        key: MsgNode(dict_to_msg_nodes(value) if isinstance(value, dict)
                     else value)
        for key, value in messages.items()
    }


def dict_from_msg_nodes(messages: MsgDict) -> PureDict:
    return {key: dict_from_msg_nodes(node.value)
            if isinstance(node.value, dict) else node.value
            for key, node in messages.items()}


def dump(messages: MsgDict, filename: str) -> None:
    if os.path.splitext(filename)[1] == ".jaml":
        with open(filename, "w", encoding=config.encoding) as f:
            f.write(jaml.dump(messages))
    else:
        messages = dict_from_msg_nodes(messages)
        with open(filename, "wb") as f:
            f.write(yaml.dump(messages, indent=4, sort_keys=False,
                              encoding="utf-8", allow_unicode=True))
